import FilterAltIcon from '@mui/icons-material/FilterAlt';
import FunctionsIcon from '@mui/icons-material/Functions';
import GridOnIcon from '@mui/icons-material/GridOn';
import NumbersIcon from '@mui/icons-material/Numbers';
import AbcIcon from '@mui/icons-material/Abc';
import { useMemo, useState } from 'react';
import { AllLogicDescriptions, AllLogicMap, Handles, NodeAttribute, QueryElementTypes, QueryGraphNodes, toHandleData } from '../../model';
import { OnConnectStartParams, XYPosition } from 'reactflow';
import { ConnectingNodeDataI } from '../utils/connectorDrop';
import { useQuerybuilderGraph, useQuerybuilderSettings, useSchemaGraph } from '@graphpolaris/shared/lib/data-access';
import { toQuerybuilderGraphology, setQuerybuilderGraphology } from '@graphpolaris/shared/lib/data-access/store/querybuilderSlice';
import { useDispatch } from 'react-redux';
import { toSchemaGraphology } from '@graphpolaris/shared/lib/data-access/store/schemaSlice';
import { SchemaAttribute, SchemaEdge, SchemaNode } from '@graphpolaris/shared/lib/schema';
import { schemaExpandRelation } from '@graphpolaris/shared/lib/schema/schema-utils';

export const QueryBuilderRelatedNodesPanel = (props: {
  reactFlowWrapper: HTMLDivElement | null;
  className?: string;
  title?: string;
  onFinished: () => void;
  connection?: ConnectingNodeDataI | null;
}) => {
  const schema = useSchemaGraph();
  const graph = useQuerybuilderGraph();
  const dispatch = useDispatch();
  const graphologyGraph = toQuerybuilderGraphology(graph);
  const schemaGraph = toSchemaGraphology(schema);
  const queryBuilderSettings = useQuerybuilderSettings();
  const [selectedType, setSelectedType] = useState<'entity' | 'relation' | 'all'>('all');

  const handleData = props.connection?.params?.handleId ? toHandleData(props.connection.params.handleId) : null;

  const relatedEntities = useMemo<SchemaNode[]>(() => {
    if (!props.connection) return [];
    const type = props.connection.node.type;
    if (type !== QueryElementTypes.Entity && type !== QueryElementTypes.Relation)
      throw new Error('Connection is not an entity or relation');

    if (type === QueryElementTypes.Entity) {
      // find entities that are connected to the current entity
      const nodes =
        handleData?.handleType === Handles.EntityRight
          ? schemaGraph.outboundNeighbors(props.connection.node.schemaKey)
          : schemaGraph.inboundNeighbors(props.connection.node.schemaKey);
      return nodes.map((node) => schemaGraph.getNodeAttributes(node) as SchemaNode);
    } else if (type === QueryElementTypes.Relation) {
      // find relations of the current entity
      const attributes = schemaGraph.getEdgeAttributes(props.connection.node.schemaKey);
      return handleData?.handleType === Handles.RelationRight
        ? [schemaGraph.getNodeAttributes(attributes.to)]
        : [schemaGraph.getNodeAttributes(attributes.from)];
    }

    return [];
  }, [schema, graph, props]);

  const relatedRelations = useMemo<SchemaEdge[]>(() => {
    if (!props.connection) return [];
    const type = props.connection.node.type;
    if (type !== QueryElementTypes.Entity && type !== QueryElementTypes.Relation)
      throw new Error('Connection is not an entity or relation');

    if (type === QueryElementTypes.Entity) {
      // find entities on the edge of the current relation
      const edges =
        handleData?.handleType === Handles.EntityRight
          ? schemaGraph.outboundEdges(props.connection.node.schemaKey)
          : schemaGraph.inboundEdges(props.connection.node.schemaKey);
      return edges.map((edge) => ({ ...schemaGraph.getEdgeAttributes(edge), label: edge } as SchemaEdge));
    } else if (type === QueryElementTypes.Relation) {
      // find relations that are connected to the proper neighboring entity of current relation
      const attributes = schemaGraph.getEdgeAttributes(props.connection.node.schemaKey);
      const edges =
        handleData?.handleType === Handles.RelationRight
          ? schemaGraph.outboundEdges(attributes.to)
          : schemaGraph.inboundEdges(attributes.from);
      return edges.map((edge) => ({ ...schemaGraph.getEdgeAttributes(edge), label: edge } as SchemaEdge));
    }

    return [];
  }, [schema, graph, props]);

  const newEntity = (entity: SchemaNode) => {
    if (!props.connection) {
      props.onFinished();
      return;
    }

    const params = props.connection.params;
    const position = props.connection.position;

    const newNode = graphologyGraph.addPill2Graphology(
      {
        type: QueryElementTypes.Entity,
        x: position.x,
        y: position.y,
        name: entity.name,
        schemaKey: entity.name,
      },
      schemaGraph.getNodeAttribute(entity.name, 'attributes')
    );

    if (!newNode?.id) throw new Error('Logic node has no id');
    if (!newNode?.name) throw new Error('Logic node has no name');
    if (!params.handleId) throw new Error('Connection has no source or target');

    if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
      graphologyGraph.addEdge2Graphology(graphologyGraph.getNodeAttributes(params.nodeId), newNode);
    else graphologyGraph.addEdge2Graphology(newNode, graphologyGraph.getNodeAttributes(params.nodeId));

    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onFinished();
  };
  const newRelation = (relation: SchemaEdge) => {
    if (!props.connection) {
      props.onFinished();
      return;
    }

    const params = props.connection.params;
    const position = props.connection.position;

    const newNode = graphologyGraph.addPill2Graphology(
      {
        type: QueryElementTypes.Relation,
        x: position.x,
        y: position.y,
        schemaKey: relation.label,
        depth: { min: queryBuilderSettings.depth.min, max: queryBuilderSettings.depth.max },
        name: relation.collection,
        collection: relation.collection,
      },
      schemaGraph.getEdgeAttribute(relation.label, 'attributes')
    );

    if (!newNode?.id) throw new Error('Logic node has no id');
    if (!newNode?.name) throw new Error('Logic node has no name');
    if (!params.handleId) throw new Error('Connection has no source or target');

    if (handleData?.handleType === Handles.EntityRight || handleData?.handleType === Handles.RelationRight)
      graphologyGraph.addEdge2Graphology(graphologyGraph.getNodeAttributes(params.nodeId), newNode);
    else graphologyGraph.addEdge2Graphology(newNode, graphologyGraph.getNodeAttributes(params.nodeId));
    dispatch(setQuerybuilderGraphology(graphologyGraph));
    props.onFinished();
  };

  return (
    <div className={props.className + ' card'}>
      {props.title && <h1 className="card-title mb-7">{props.title}</h1>}
      <div className="overflow-x-hidden h-[75rem] w-full gap-2 pt-3">
        <div className="w-full">
          <button
            className={'btn w-[calc(50%-0.25rem)] mr-2 normal-case text-lg mb-3 ' + (selectedType !== 'relation' ? 'btn-active' : '')}
            onClick={(e) => {
              e.preventDefault();
              selectedType === 'entity' ? setSelectedType('all') : setSelectedType('entity');
            }}
          >
            Related Entities
          </button>
          <button
            className={'btn w-[calc(50%-0.25rem)] text-lg normal-case mb-3 ' + (selectedType !== 'entity' ? 'btn-active' : '')}
            onClick={(e) => {
              e.preventDefault();
              selectedType === 'relation' ? setSelectedType('all') : setSelectedType('relation');
            }}
          >
            Related Relations
          </button>
        </div>
        <div className="w-full flex flex-row gap-2">
          {selectedType !== 'relation' && (
            <ul className="menu p-0 [&_li>*]:rounded-none w-full pb-10">
              {relatedEntities.map((entity) => {
                return (
                  <button className="btn btn-sm border-entity-400 normal-case" key={entity.name} onClick={() => newEntity(entity)}>
                    {entity.name}
                  </button>
                );
              })}
            </ul>
          )}
          {selectedType !== 'entity' && (
            <ul className="menu p-0 [&_li>*]:rounded-none w-full pb-10">
              {relatedRelations.map((relation) => {
                return (
                  <button className="btn btn-sm border-relation-400 normal-case" key={relation.name} onClick={() => newRelation(relation)}>
                    {relation.name}
                  </button>
                );
              })}
            </ul>
          )}
        </div>
      </div>
    </div>
  );
};
