import { AlgorithmToLayoutProvider, AllLayoutAlgorithms, LayoutFactory } from '@graphpolaris/shared/lib/graph-layout';
import { schemaGraphology2Reactflow, schemaExpandRelation } from '@graphpolaris/shared/lib/schema/schema-utils';
import { useSchemaGraph, useSchemaSettings, useSearchResultSchema, useSessionCache } from '@graphpolaris/shared/lib/data-access/store';
import { SmartBezierEdge, SmartStepEdge, SmartStraightEdge } from '@tisoap/react-flow-smart-edge';

import { useEffect, useMemo, useRef, useState } from 'react';
import ReactFlow, { Node, Edge, ReactFlowProvider, useNodesState, useEdgesState, ReactFlowInstance } from 'reactflow';

import 'reactflow/dist/style.css';

import { ConnectionDragLine, ConnectionLine } from '@graphpolaris/shared/lib/querybuilder/pills';
import { EntityNode } from '../pills/nodes/entity/entity-node';
import { RelationNode } from '../pills/nodes/relation/relation-node';
import NodeEdge from '../pills/edges/node-edge';
import SelfEdge from '../pills/edges/self-edge';
import { useSchemaAPI } from '../../data-access';
import { SchemaDialog } from './schemaDialog';
import { toSchemaGraphology } from '../../data-access/store/schemaSlice';
import { Button } from '../../components/buttons';
import ControlContainer from '../../components/controls';

interface Props {
  content?: string;
  auth?: boolean;
}

const onInit = (reactFlowInstance: ReactFlowInstance) => {
  setTimeout(() => reactFlowInstance.fitView(), 100);
};

const nodeTypes = {
  entity: EntityNode,
  relation: RelationNode,
};
const edgeTypes = {
  nodeEdge: NodeEdge,
  selfEdge: SelfEdge,
  bezier: SmartBezierEdge,
  connection: ConnectionLine,
  straight: SmartStraightEdge,
  step: SmartStepEdge,
};

export const Schema = (props: Props) => {
  const api_schema = useSchemaAPI();
  const session = useSessionCache();
  const settings = useSchemaSettings();
  const searchResults = useSearchResultSchema();
  const [toggleSchemaSettings, setToggleSchemaSettings] = useState(false);
  const [nodes, setNodes, onNodeChanged] = useNodesState([] as Node[]);
  const [edges, setEdges, onEdgeChanged] = useEdgesState([] as Edge[]);
  const [firstUserConnection, setFirstUserConnection] = useState<boolean>(true);
  const [auth, setAuth] = useState(props.auth);

  const reactFlowInstanceRef = useRef<ReactFlowInstance | null>(null);

  // In case the schema is updated
  const schemaGraph = useSchemaGraph();
  const schemaGraphology = useMemo(() => toSchemaGraphology(schemaGraph), [schemaGraph]);
  const layout = useRef<AlgorithmToLayoutProvider<AllLayoutAlgorithms>>();

  function updateLayout() {
    const layoutFactory = new LayoutFactory();
    layout.current = layoutFactory.createLayout(settings.layoutName);
  }

  const fitView = () => {
    if (reactFlowInstanceRef.current) {
      reactFlowInstanceRef.current.fitView();
    }
  };

  useEffect(() => {
    updateLayout();
    sessionStorage.setItem('firstUserConnection', firstUserConnection.toString());
  }, []);

  useEffect(() => {
    setAuth(props.auth);
  }, [props.auth]);

  useEffect(() => {
    if (schemaGraphology === undefined || schemaGraphology.order == 0) {
      setNodes([]);
      setEdges([]);
      return;
    }

    updateLayout();
    const expandedSchema = schemaExpandRelation(schemaGraphology);
    layout.current?.layout(expandedSchema);
    const schemaFlow = schemaGraphology2Reactflow(expandedSchema, settings.connectionType);
    setNodes(schemaFlow.nodes);
    setEdges(schemaFlow.edges);
  }, [schemaGraph, settings]);

  useEffect(() => {
    setNodes((nds) =>
      nds.map((node) => ({
        ...node,
        selected: searchResults.includes(node.id) || searchResults.includes(node.data.label),
      })),
    );
  }, [searchResults]);

  return (
    <div className="schema-panel w-full h-full">
      <SchemaDialog open={toggleSchemaSettings} onClose={() => setToggleSchemaSettings(false)} />
      <div className="relative flex items-center justify-between z-[2] py-0 px-2 bg-secondary-100  border-b border-secondary-200">
        <h1 className="text-xs font-semibold text-secondary-800">Schema</h1>
        <ControlContainer>
          <Button type="secondary" variant="ghost" size="xs" iconName="Fullscreen" onClick={fitView} />
          <Button
            type="secondary"
            variant="ghost"
            size="xs"
            iconName="Cached"
            onClick={(e) => {
              e.stopPropagation();
              api_schema.RequestSchema(session.currentDatabase);
            }}
          />
          <Button
            type="secondary"
            variant="ghost"
            size="xs"
            iconName="Settings"
            onClick={(e) => {
              e.stopPropagation();
              setToggleSchemaSettings(!toggleSchemaSettings);
            }}
          />
        </ControlContainer>
      </div>
      {nodes.length === 0 ? (
        <p className="text-sm">No Elements</p>
      ) : (
        <ReactFlowProvider>
          <div className="h-[calc(100%-.8rem)] w-full">
            <ReactFlow
              snapGrid={[10, 10]}
              snapToGrid
              onlyRenderVisibleElements={false}
              nodesDraggable={true}
              nodeTypes={nodeTypes}
              edgeTypes={edgeTypes}
              connectionLineComponent={ConnectionDragLine}
              onNodesChange={onNodeChanged}
              onEdgesChange={onEdgeChanged}
              nodes={nodes}
              edges={edges}
              onInit={(reactFlowInstance) => {
                reactFlowInstanceRef.current = reactFlowInstance;
                onInit(reactFlowInstance);
              }}
              proOptions={{ hideAttribution: true }}
            ></ReactFlow>
          </div>
        </ReactFlowProvider>
      )}
    </div>
  );
};

export default Schema;
